"""A source fetching currency prices from OANDA.

Valid tickers are in the form "XXX_YYY", such as "EUR_USD".

Here is the API documentation:
http://developer.oanda.com/rest-live/rates/

For example:
https://api-fxtrade.oanda.com/v1/candles?instrument=EUR_USD&granularity=D&start=2016-03-27T00%3A00%3A00Z&end=2016-04-04T00%3A00%3A00Z&candleFormat=midpoint
"""
__author__ = "Martin Blais <blais@furius.ca>"

import re
import datetime
import json
import logging
from urllib import parse

from dateutil import tz

from beancount.core.number import D
from beancount.prices import source
from beancount.utils import net_utils


URL = "https://api-fxtrade.oanda.com/v1/candles"


def _get_currencies(ticker):
    """Parse the base and quote currencies from the ticker.

    Args:
      ticker: A string, the symbol in XXX_YYY format.
    Returns:
      A pair of (base, quote) currencies.
    """
    match = re.match("([A-Z]+)_([A-Z]+)$", ticker)
    if not match:
        return None, None
    return match.groups()


def _fetch_candles(params):
    """Fetch the given URL from OANDA and return a list of (utc-time, price).

    Args:
      params: A dict of URL params values.
    Returns:
      A list of (time, price) points.
    """

    url = '?'.join((URL, parse.urlencode(sorted(params.items()))))
    logging.info("Fetching '%s'", url)

    # Fetch the data.
    response = net_utils.retrying_urlopen(url)
    if response is None:
        return None
    data_string = response.read().decode('utf-8')

    # Parse it.
    data = json.loads(data_string, parse_float=D)
    try:
        # Find the candle with the latest time before the given time we're searching
        # for.
        time_prices = []
        candles = sorted(data['candles'], key=lambda candle: candle['time'])
        for candle in candles:
            candle_dt_utc = datetime.datetime.strptime(
                candle['time'], r"%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=tz.tzutc())
            candle_price = D(candle['openMid'])
            time_prices.append((candle_dt_utc, candle_price))
    except KeyError:
        logging.error("Unexpected response data: %s", data)
        return None
    return time_prices


class Source(source.Source):
    "OANDA price source extractor."

    def get_latest_price(self, ticker):
        """See contract in beancount.prices.source.Source."""

        _, quote_currency = _get_currencies(ticker)
        if quote_currency is None:
            logging.error("Invalid price source ticker '%s'; must be like 'EUR_USD'",
                          ticker)
            return

        # Query at the current (latest) time.
        params_dict = {
            'instrument': ticker,
            'granularity': 'S5',  # Every two hours.
            'count': '10',
            'candleFormat': 'midpoint',
        }
        time_prices = _fetch_candles(params_dict)
        if time_prices is None:
            logging.error("No prices returned")
            return

        # Get the latest price point available.
        time, price = sorted(time_prices)[-1]

        # Use current local date as the date.
        current_date = datetime.datetime.now()

        return source.SourcePrice(price, current_date, quote_currency)

    def get_historical_price(self, ticker, date):
        """See contract in beancount.prices.source.Source."""

        _, quote_currency = _get_currencies(ticker)
        if quote_currency is None:
            logging.error("Invalid price source ticker '%s'; must be like 'EUR_USD'",
                          ticker)
            return

        # Find the boundary dates to query in UTC timezone.
        start_utc = datetime.datetime(date.year, date.month, date.day, 0, 0, 0,
                                      tzinfo=tz.tzlocal()).astimezone(tz.tzutc())
        end_utc = start_utc + datetime.timedelta(days=1)

        interval_begin_utc = (start_utc - datetime.timedelta(days=5))
        interval_end_utc = (end_utc + datetime.timedelta(days=5))

        # Build the query.
        params_dict = {
            'instrument': ticker,
            'granularity': 'H2',  # Every two hours.
            'candleFormat': 'midpoint',
            'start': interval_begin_utc.isoformat('T'),
            'end': interval_end_utc.isoformat('T'),
        }
        time_prices = _fetch_candles(params_dict)
        if time_prices is None:
            logging.error("No prices returned")
            return

        # Get all the prices with the same date.
        same_date = [item
                     for item in time_prices
                     if start_utc <= item[0] < end_utc]
        if same_date:
            # Find the min/max and return the median of all prices.
            sorted_prices = sorted(same_date, key=lambda item: item[1])
            time, price = sorted_prices[len(sorted_prices)//2]
        else:
            # No price matching the date were found; use the midpoint of the
            # last price before the day interval and the first price after the
            # day interval.
            before_time, before_price = min(item
                                            for item in time_prices
                                            if item[0] < start_utc)
            after_time, after_price = max(item
                                          for item in time_prices
                                          if item[0] >= end_utc)
            price = (after_price + before_price) / 2
            time = before_time + (after_time - before_time)/2

        return source.SourcePrice(price, time, quote_currency)



"""Algorithms for 'booking' inventory, that is, the process of finding a
matching lot when reducing the content of an inventory.
"""
__copyright__ = "Copyright (C) 2015-2017  Martin Blais"
__license__ = "GNU GPLv2"

import collections
import copy

from beancount.core.amount import Amount
from beancount.core.data import Posting
from beancount.core.data import Transaction
from beancount.core.inventory import Inventory
from beancount.core.number import MISSING
from beancount.core.number import ZERO
from beancount.core.position import Cost
from beancount.core.position import Position
from beancount.core import convert

from beancount.core import account
from beancount.core import interpolate



__sanity_checks__ = False


SimpleBookingError = collections.namedtuple('SimpleBookingError', 'source message entry')


def book(entries, options_map, unused_booking_methods):
    """Run a local interpolation on a list of incomplete entries from the parser.

    Note: this does not take previous positions into account.

    !WARNING!!! This destructively modifies some of the Transaction entries directly.

    Args:
      incomplete_entries: A list of directives, with some postings possibly left
        with incomplete amounts as produced by the parser.
      options_map: An options dict as produced by the parser.
      unused_booking_methods: (Unused.)
    Returns:
      A pair of
        entries: A list of interpolated entries with all their postings completed.
        errors: New errors produced during interpolation.
    """
    # Perform simple booking, that is convert the CostSpec instances to Cost,
    # not even looking at inventory contents.
    entries_with_lots, errors = convert_lot_specs_to_lots(entries)

    new_entries = []
    for entry in entries_with_lots:
        if isinstance(entry, Transaction):

            # Check that incompleteness may only occur at the posting level; reject
            # otherwise. These cases are handled by the FULL booking algorithm but
            # not so well by the SIMPLE algorithm. The new parsing is more liberal
            # than this old code can handle, so explicitly reject cases where it
            # would fail.
            skip = False
            for posting in entry.postings:
                units = posting.units
                if units is not MISSING:
                    if units.number is MISSING or units.currency is MISSING:
                        errors.append(SimpleBookingError(
                            entry.meta,
                            "Missing number or currency on units not handled",
                            None))
                        skip = True
                        break

                price = posting.price
                if price is not None:
                    if price.number is MISSING or price.currency is MISSING:
                        errors.append(SimpleBookingError(
                            entry.meta,
                            "Missing number or currency on price not handled",
                            None))
                        skip = True
                        break
            if skip:
                continue

            # Balance incomplete auto-postings and set the parent link to this
            # entry as well.
            balance_errors = balance_incomplete_postings(entry, options_map)
            if balance_errors:
                errors.extend(balance_errors)

            # Check that the balance actually is empty.
            if __sanity_checks__:
                residual = interpolate.compute_residual(entry.postings)
                tolerances = interpolate.infer_tolerances(entry.postings, options_map)
                assert residual.is_small(tolerances), "Invalid residual {}".format(residual)

        new_entries.append(entry)

    return new_entries, errors


def convert_lot_specs_to_lots(entries):
    """For all the entries, convert the posting's position's CostSpec to Cost
    instances. In the simple method, the data provided in the CostSpec must
    unambiguously provide a way to compute the cost amount.

    This essentially replicates the way the old parser used to work, but
    allowing positions to have the fuzzy lot specifications instead of the
    resolved ones. We used to simply compute the costs locally, and this gets
    rid of the CostSpec to produce the Cost without fuzzy matching. This is only
    there for the sake of transition to the new matching logic.

    Args:
      entries: A list of incomplete directives as per the parser.
    Returns:
      A list of entries whose postings's position costs have been converted to
      Cost instances but that may still be incomplete.
    Raises:
      ValueError: If there's a unacceptable number.
    """
    new_entries = []
    errors = []
    for entry in entries:
        if not isinstance(entry, Transaction):
            new_entries.append(entry)
            continue

        new_postings = []
        for posting in entry.postings:
            try:
                units = posting.units
                cost_spec = posting.cost
                cost = convert_spec_to_cost(units, cost_spec)
                if cost_spec is not None and cost is None:
                    errors.append(
                        SimpleBookingError(entry.meta,
                                           "Cost syntax not supported; cost spec ignored",
                                           None))

                if cost and isinstance(units, Amount):
                    # If there is a cost, we don't allow either a cost value of
                    # zero, nor a zero number of units. Note that we allow a price
                    # of zero as the only special case (for conversion entries), but
                    # never for costs.
                    if units.number == ZERO:
                        raise ValueError('Amount is zero: "{}"'.format(units))
                    if cost.number is not None and cost.number < ZERO:
                        raise ValueError('Cost is negative: "{}"'.format(cost))
            except ValueError as exc:
                errors.append(SimpleBookingError(entry.meta, str(exc), None))
                cost = None
            new_postings.append(posting._replace(cost=cost))
        new_entries.append(entry._replace(postings=new_postings))
    return new_entries, errors


def convert_spec_to_cost(units, cost_spec):
    """Convert a posting's CostSpec instance to a Cost.

    Args:
      units: An instance of Amount.
      cost_spec: An instance of CostSpec.
    Returns:
      An instance of Cost.
    """
    cost = cost_spec
    errors = []
    if isinstance(units, Amount):
        currency = units.currency
        if cost_spec is not None:
            number_per, number_total, cost_currency, date, label, merge = cost_spec

            # Compute the cost.
            if number_per is not MISSING or number_total is not None:
                if number_total is not None:
                    # Compute the per-unit cost if there is some total cost
                    # component involved.
                    units_num = units.number
                    cost_total = number_total
                    if number_per is not MISSING:
                        cost_total += number_per * units_num
                    unit_cost = cost_total / abs(units_num)
                else:
                    unit_cost = number_per
                cost = Cost(unit_cost, cost_currency, date, label)
            else:
                cost = None
    return cost


def get_incomplete_postings(entry, options_map):
    """Balance an entry with auto-postings and return an updated list of completed postings.

    Returns a new list of balanced postings, with the incomplete postings
    replaced with completed ones. This is probably the only place where there
    is a bit of non-trivial logic in this entire project (and the rewrite was
    to make sure it was *that* simple.)

    Note that inferred postings are tagged via metatada with an '__automatic__'
    field added to them with a true boolean value.

    Note: The 'postings' parameter may be modified or destroyed for performance
    reasons; don't reuse it.

    Args:
      entry: An instance of a valid directive.
      options_map: A dict of options, as produced by the parser.
    Returns:
      A tuple of:
        postings: a list of new postings to replace the entry's unbalanced
          postings.
        inserted: A boolean set to true if we've inserted new postings.
        errors: A list of balance errors generated during the balancing process.
        residual: A Inventory instance, the residual amounts after balancing
          the postings.
        tolerances: The tolerances inferred in the process, using the postings
          provided.
    """
    # Make a copy of the original list of postings.
    postings = list(entry.postings)

    # Errors during balancing.
    balance_errors = []

    # The list of postings without and with an explicit position.
    auto_postings_indices = []

    # Currencies seen in complete postings.
    currencies = set()

    # An inventory to accumulate the residual balance.
    residual = Inventory()

    # A dict of values for default tolerances.
    tolerances = interpolate.infer_tolerances(postings, options_map)

    # Process all the postings.
    has_nonzero_amount = False
    has_regular_postings = False
    for i, posting in enumerate(postings):
        units = posting.units
        if units is MISSING or units is None:
            # This posting will have to get auto-completed.
            auto_postings_indices.append(i)
        else:
            currencies.add(units.currency)

            # Compute the amount to balance and update the inventory.
            weight = convert.get_weight(posting)
            residual.add_amount(weight)

            has_regular_postings = True
            if weight:
                has_nonzero_amount = True

    # If there are auto-postings, fill them in.
    has_inserted = False
    if auto_postings_indices:
        # If there are too many such postings, we can't do anything, barf.
        if len(auto_postings_indices) > 1:
            balance_errors.append(
                SimpleBookingError(entry.meta,
                                   "Too many auto-postings; cannot fill in",
                                   entry))
            # Delete the redundant auto-postings.
            for index in sorted(auto_postings_indices[1:], reverse=1):
                del postings[index]

        index = auto_postings_indices[0]
        old_posting = postings[index]
        assert old_posting.price is None

        residual_positions = residual.get_positions()

        # If there are no residual positions, we want to still insert a posting
        # but with a zero position for each currency, so that the posting shows
        # up anyhow. We insert one such posting for each currency seen in the
        # complete postings. Note: if all the non-auto postings are zero, we
        # want to avoid sending a warning; the input text clearly implies the
        # author knows this would be useless.
        new_postings = []
        if not residual_positions and ((has_regular_postings and has_nonzero_amount) or
                                       not has_regular_postings):
            balance_errors.append(
                SimpleBookingError(entry.meta,
                                   "Useless auto-posting: {}".format(residual), entry))
            for currency in currencies:
                units = Amount(ZERO, currency)
                meta = copy.copy(old_posting.meta) if old_posting.meta else {}
                meta[interpolate.AUTOMATIC_META] = True
                new_postings.append(
                    Posting(old_posting.account, units, None,
                            None, old_posting.flag, old_posting.meta))
                has_inserted = True
        else:
            # Convert all the residual positions in inventory into a posting for
            # each position.
            for pos in residual_positions:
                pos = -pos

                units = pos.units
                new_units = Amount(
                    interpolate.quantize_with_tolerance(tolerances,
                                                        units.currency,
                                                        units.number),
                    units.currency)

                meta = copy.copy(old_posting.meta) if old_posting.meta else {}
                meta[interpolate.AUTOMATIC_META] = True
                new_postings.append(
                    Posting(old_posting.account, new_units, pos.cost,
                            None, old_posting.flag, meta))
                has_inserted = True

                # Note/FIXME: This is dumb; refactor cost computation so we can
                # reuse it directly.
                new_pos = Position(new_units, pos.cost)

                # Update the residuals inventory.
                weight = convert.get_cost(new_pos)
                residual.add_amount(weight)

        postings[index:index+1] = new_postings

    else:
        # Checking for unbalancing transactions has been moved to the validation
        # stage, so although we already have the input transaction's residuals
        # conveniently precomputed here, we are postponing the check to allow
        # plugins to "fixup" unbalancing transactions. We want to allow users to
        # be able to input unbalancing transactions as long as the final
        # transactions objects that appear on the stream (after processing the
        # plugins) are balanced. See {9e6c14b51a59}.
        pass

    return (postings, has_inserted, balance_errors, residual, tolerances)


def balance_incomplete_postings(entry, options_map):
    """Balance an entry with incomplete postings, modifying the
    empty postings on the entry itself. This sets the parent of
    all the postings to this entry. Futhermore, it stores the dict
    of inferred tolerances as metadata.

    WARNING: This destructively modifies entry itself!

    Args:
      entry: An instance of a valid directive. This entry is modified by
        having new postings inserted to it.
      options_map: A dict of options, as produced by the parser.
    Returns:
      A list of errors, or None, if none occurred.
    """
    # No postings... nothing to do.
    if not entry.postings:
        return None

    # Get the list of corrected postings.
    (postings, unused_inserted, errors,
     residual, tolerances) = get_incomplete_postings(entry, options_map)

    # If we need to accumulate rounding error to accumulate the residual, add
    # suitable postings here.
    if not residual.is_empty():
        rounding_subaccount = options_map["account_rounding"]
        if rounding_subaccount:
            account_rounding = account.join(options_map['name_equity'], rounding_subaccount)
            rounding_postings = interpolate.get_residual_postings(residual,
                                                                  account_rounding)
            postings.extend(rounding_postings)

    # If we could make this faster to avoid the unnecessary copying, it would
    # make parsing substantially faster.
    entry.postings.clear()
    for posting in postings:
        entry.postings.append(posting)

    if entry.meta is None:
        entry.meta = {}
    entry.meta[interpolate.AUTOMATIC_TOLERANCES] = tolerances

    return errors or None



"""CSV importer.
"""
__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import csv
import datetime
import enum
import io
from os import path

from beancount.core.number import D
from beancount.core.amount import Amount
from beancount.utils.date_utils import parse_date_liberally
from beancount.core import data
from beancount.ingest import importer
from beancount.ingest.importers import regexp


# The set of interpretable columns.
class Col(enum.Enum):
    # The settlement date, the date we should create the posting at.
    DATE = '[DATE]'

    # The date at which the transaction took place.
    TXN_DATE = '[TXN_DATE]'

    # The payee field.
    PAYEE = '[PAYEE]'

    # The narration fields. Use multiple fields to combine them together.
    NARRATION = NARRATION1 = '[NARRATION1]'
    NARRATION2 = '[NARRATION2]'
    NARRATION3 = '[NARRATION3]'

    # The amount being posted.
    AMOUNT = '[AMOUNT]'

    # Debits and credits being posted in separate, dedicated columns.
    AMOUNT_DEBIT = '[DEBIT]'
    AMOUNT_CREDIT = '[CREDIT]'

    # The balance amount, after the row has posted.
    BALANCE = '[BALANCE]'

    # A field to use as a tag name.
    TAG = '[TAG]'

    # A column which says DEBIT or CREDIT (generally ignored).
    DRCR = '[DRCR]'


def get_amounts(iconfig, row):
    """Get the amount columns of a row.

    Args:
      iconfig: A dict of Col to row index.
      row: A row array containing the values of the given row.
    Returns:
      A pair of (debit-amount, credit-amount), both of which are either an
      instance of Decimal or None, or not available.
    """
    debit, credit = None, None
    if Col.AMOUNT in iconfig:
        credit = row[iconfig[Col.AMOUNT]]
    else:
        debit, credit = [row[iconfig[col]] if col in iconfig else None
                         for col in [Col.AMOUNT_DEBIT, Col.AMOUNT_CREDIT]]
    return (-D(debit) if debit else None,
            D(credit) if credit else None)


class Importer(regexp.RegexpImporterMixin, importer.ImporterProtocol):
    """Importer for Chase credit card accounts."""

    def __init__(self, config, account, currency, regexps,
                 institution=None,
                 debug=False):
        """Constructor.

        Args:
          config: A dict of Col enum types to the names or indexes of the columns.
          account: An account string, the account to post this to.
          currency: A currency string, the currenty of this account.
          regexps: A list of regular expression strings.
          institution: An optional name of an institution to rename the files to.
          header:
        """
        if isinstance(regexps, str):
            regexps = [regexps]
        assert isinstance(regexps, list)
        regexp.RegexpImporterMixin.__init__(self, regexps)

        assert isinstance(config, dict)
        self.config = config

        self.account = account
        self.currency = currency
        self.debug = debug

        # FIXME: This probably belongs to a mixin, not here.
        self.institution = institution

    def name(self):
        name = self.name or super().name()
        return '{}: "{}"'.format(super().name(), self.file_account(None))

    def file_account(self, _):
        return self.account

    def file_name(self, file):
        filename = path.splitext(path.basename(file.name))[0]
        if self.institution:
            filename = '{}.{}'.format(self.institution, filename)
        return '{}.csv'.format(filename)

    def file_date(self, file):
        "Get the maximum date from the file."
        iconfig, has_header = normalize_config(self.config, file.head())
        if Col.DATE in iconfig:
            reader = iter(csv.reader(open(file.name)))
            if has_header:
                next(reader)
            max_date = None
            for row in reader:
                if not row:
                    continue
                date_str = row[iconfig[Col.DATE]]
                date = parse_date_liberally(date_str)
                if max_date is None or date > max_date:
                    max_date = date
            return max_date


    # def get_description(self, row):
    #     """Extract the payee and narration from the row.

    #     This is a place where you can customize and combine multiple fields
    #     together.

    #     Args:
    #       row: A collections.namedtuple object representing the row.
    #     Returns:
    #       A pair of (payee, narration) string, either of which may be None.
    #     """
    #     payee = ({getattr(row, self.config[Col.PAYEE])}
    #              if Col.PAYEE in self.config else
    #              None)
    #     narration = ({getattr(row, self.config[Col.NARRATION])}
    #                  if Col.NARRATION in self.config else
    #                  None)
    #     return payee, narration

    def extract(self, file):
        entries = []

        # Normalize the configuration to fetch by index.
        iconfig, has_header = normalize_config(self.config, file.head())

        # Skip header, if one was detected.
        reader = iter(csv.reader(open(file.name)))
        if has_header:
            next(reader)
        def get(row, ftype):
            return row[iconfig[ftype]] if ftype in iconfig else None

        # Parse all the transactions.
        first_row = last_row = None
        for index, row in enumerate(reader, 1):
            if not row:
                continue

            # If debugging, print out the rows.
            if self.debug: print(row)

            if first_row is None:
                first_row = row
            last_row = row

            # Extract the data we need from the row, based on the configuration.
            date = get(row, Col.DATE)
            txn_date = get(row, Col.TXN_DATE)

            payee = get(row, Col.PAYEE)
            fields = filter(None, [get(row, field)
                                   for field in (Col.NARRATION1,
                                                 Col.NARRATION2,
                                                 Col.NARRATION3)])
            narration = ' -- '.join(fields)

            tag = get(row, Col.TAG)
            tags = {tag} if tag is not None else data.EMPTY_SET

            # Create a transaction and add it to the list of new entries.
            meta = data.new_metadata(file.name, index)
            if txn_date is not None:
                meta['txndate'] = parse_date_liberally(txn_date)
            date = parse_date_liberally(date)
            txn = data.Transaction(meta, date, self.FLAG, payee, narration,
                                   tags, data.EMPTY_SET, [])
            entries.append(txn)

            amount_debit, amount_credit = get_amounts(iconfig, row)
            for amount in [amount_debit, amount_credit]:
                if amount is None:
                    continue
                units = Amount(amount, self.currency)
                txn.postings.append(
                    data.Posting(self.account, units, None, None, None, None))

        # Parse the final balance.
        if Col.BALANCE in iconfig and first_row and last_row:
            # Figure out if the file is in ascending or descending order.
            first_date = parse_date_liberally(get(first_row, Col.DATE))
            last_date = parse_date_liberally(get(last_row, Col.DATE))
            is_ascending = first_date < last_date

            # Choose between the first or the last row based on the date.
            row = last_row if is_ascending else first_row
            date = parse_date_liberally(get(row, Col.DATE)) + datetime.timedelta(days=1)
            balance = D(get(row, Col.BALANCE))
            meta = data.new_metadata(file.name, index)
            entries.append(
                data.Balance(meta, date,
                             self.account, Amount(balance, self.currency),
                             None, None))

        return entries


def normalize_config(config, head):
    """Using the header line, convert the configuration field name lookups to int indexes.

    Args:
      config: A dict of Col types to string or indexes.
      head: A string, some decent number of bytes of the head of the file.
    Returns:
      A pair of
        A dict of Col types to integer indexes of the fields, and
        a boolean, true if the file has a header.
    Raises:
      ValueError: If there is no header and the configuration does not consist
        entirely of integer indexes.
    """
    has_header = csv.Sniffer().has_header(head)
    if has_header:
        header = next(csv.reader(io.StringIO(head)))
        field_map = {field_name.strip(): index
                     for index, field_name in enumerate(header)}
        index_config = {}
        for field_type, field in config.items():
            if isinstance(field, str):
                field = field_map[field]
            index_config[field_type] = field
    else:
        if any(not isinstance(field, int)
               for field_type, field in config.items()):
            raise ValueError("CSV config without header has non-index fields: "
                             "{}".format(config))
        index_config = config
    return index_config, has_header



__copyright__ = "Copyright (C) 2015-2016  Martin Blais"
__license__ = "GNU GPLv2"

from beancount.core.number import D
from beancount.core.data import create_simple_posting as P

from beancount import loader
from beancount.core import data
from beancount.core import interpolate
from beancount.core import inventory
from beancount.core import position
from beancount.parser import cmptest
from beancount.parser import parser
from beancount.parser import booking_simple


# True if errors are generated on residual by get_incomplete_postings().
ERRORS_ON_RESIDUAL = False


class TestGetIncompletePostings(cmptest.TestCase):

    OPTIONS_MAP = {'inferred_tolerance_default': {},
                   'inferred_tolerance_multiplier': D('0.5'),
                   'account_rounding': None,
                   'infer_tolerance_from_cost': False}

    def test_get_incomplete_postings_pathological(self):
        meta = data.new_metadata(__file__, 0)

        # Test with no entries.
        entry = data.Transaction(meta, None, None, None, None,
                                 data.EMPTY_SET, data.EMPTY_SET, [])
        new_postings, has_inserted, errors, _, __ = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertFalse(has_inserted)
        self.assertEqual(0, len(new_postings))
        self.assertEqual(0, len(errors))

        # Test with only a single leg (and check that it does not balance).
        entry = data.Transaction(meta, None, None, None, None,
                                 data.EMPTY_SET, data.EMPTY_SET, [
            P(None, "Assets:Bank:Checking", "105.50", "USD"),
            ])
        (new_postings, has_inserted, errors,
         residual, tolerances) = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertFalse(has_inserted)
        self.assertEqual(1, len(new_postings))
        self.assertEqual(1 if ERRORS_ON_RESIDUAL else 0, len(errors))
        self.assertIsInstance(tolerances, dict)

        # Test with two legs that balance.
        entry = data.Transaction(
            meta, None, None, None, None, data.EMPTY_SET, data.EMPTY_SET, [
                P(None, "Assets:Bank:Checking", "105.50", "USD"),
                P(None, "Assets:Bank:Savings", "-105.50", "USD"),
            ])
        new_postings, has_inserted, errors, _, __ = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertFalse(has_inserted)
        self.assertEqual(2, len(new_postings))
        self.assertEqual(0, len(errors))

        # Test with two legs that do not balance.
        entry = data.Transaction(
            meta, None, None, None, None, data.EMPTY_SET, data.EMPTY_SET, [
                P(None, "Assets:Bank:Checking", "105.50", "USD"),
                P(None, "Assets:Bank:Savings", "-115.50", "USD"),
            ])
        new_postings, has_inserted, errors, _, __ = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertFalse(has_inserted)
        self.assertEqual(2, len(new_postings))
        self.assertEqual(1 if ERRORS_ON_RESIDUAL else 0, len(errors))

        # Test with only one auto-posting.
        entry = data.Transaction(
            meta, None, None, None, None, data.EMPTY_SET, data.EMPTY_SET, [
                P(None, "Assets:Bank:Checking", None, None),
            ])
        new_postings, has_inserted, errors, _, __ = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertFalse(has_inserted)
        self.assertEqual(0, len(new_postings))
        self.assertEqual(1, len(errors))

        # Test with an auto-posting where there is no residual.
        entry = data.Transaction(
            meta, None, None, None, None, data.EMPTY_SET, data.EMPTY_SET, [
                P(None, "Assets:Bank:Checking", "105.50", "USD"),
                P(None, "Assets:Bank:Savings", "-105.50", "USD"),
                P(None, "Assets:Bank:Balancing", None, None),
            ])
        new_postings, has_inserted, errors, _, __ = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertTrue(has_inserted)
        self.assertEqual(3, len(new_postings))
        self.assertEqual(1, len(errors))

        # Test with too many empty postings.
        entry = data.Transaction(
            meta, None, None, None, None, data.EMPTY_SET, data.EMPTY_SET, [
                P(None, "Assets:Bank:Checking", "105.50", "USD"),
                P(None, "Assets:Bank:Savings", "-106.50", "USD"),
                P(None, "Assets:Bank:BalancingA", None, None),
                P(None, "Assets:Bank:BalancingB", None, None),
            ])
        new_postings, has_inserted, errors, _, __ = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertTrue(has_inserted)
        self.assertEqual(3, len(new_postings))
        self.assertEqual(1, len(errors))

    def test_get_incomplete_postings_normal(self):
        meta = data.new_metadata(__file__, 0)

        # Test with a single auto-posting with a residual.
        entry = data.Transaction(
            meta, None, None, None, None, data.EMPTY_SET, data.EMPTY_SET, [
                P(None, "Assets:Bank:Checking", "105.50", "USD"),
                P(None, "Assets:Bank:Savings", "-115.50", "USD"),
                P(None, "Assets:Bank:Balancing", None, None),
            ])
        new_postings, has_inserted, errors, _, __ = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertTrue(has_inserted)
        self.assertEqual(3, len(new_postings))
        self.assertEqual(0, len(errors))
        self.assertTrue(interpolate.AUTOMATIC_META in new_postings[2].meta)

    def test_get_incomplete_postings_residual(self):
        meta = data.new_metadata(__file__, 0)

        # Test with a single auto-posting with a residual.
        entry = data.Transaction(
            meta, None, None, None, None, data.EMPTY_SET, data.EMPTY_SET, [
                P(None, "Assets:Bank:Checking", "105.50", "USD"),
                P(None, "Assets:Bank:Savings", "-115.501", "USD"),
                P(None, "Assets:Bank:Balancing", "10.00", "USD"),
            ])
        _, __, ___, residual, ____ = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertEqual(inventory.from_string('-0.001 USD'), residual)

    def test_get_residual_postings(self):
        residual = inventory.from_string('0.001 USD, -0.00002 CAD')
        account_rounding = 'Equity:RoundingError'
        postings = interpolate.get_residual_postings(residual, account_rounding)
        self.assertEqual(2, len(postings))
        self.assertEqual([
            P(None, "Equity:RoundingError", "-0.001", "USD"),
            P(None, "Equity:RoundingError", "0.00002", "CAD"),
            ], [posting._replace(meta=None) for posting in postings])

    def test_balance_with_large_amount(self):
        meta = data.new_metadata(__file__, 0)

        # Test with a single auto-posting with a residual.
        entry = data.Transaction(
            meta, None, None, None, None, data.EMPTY_SET, data.EMPTY_SET, [
                P(None, "Income:US:Anthem:InsurancePayments", "-275.81", "USD"),
                P(None, "Income:US:Anthem:InsurancePayments", "-23738.54", "USD"),
                P(None, "Assets:Bank:Checking", "24014.45", "USD"),
            ])
        new_postings, has_inserted, errors, _, __ = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertFalse(has_inserted)
        self.assertEqual(3, len(new_postings))
        self.assertEqual(1 if ERRORS_ON_RESIDUAL else 0, len(errors))

    def test_balance_with_zero_posting(self):
        meta = data.new_metadata(__file__, 0)
        entry = data.Transaction(
            meta, None, None, None, None, data.EMPTY_SET, data.EMPTY_SET, [
                P(None, "Income:US:Anthem:InsurancePayments", "0", "USD"),
                P(None, "Income:US:Anthem:InsurancePayments", None, None),
            ])
        new_postings, has_inserted, errors, _, __ = booking_simple.get_incomplete_postings(
            entry, self.OPTIONS_MAP)
        self.assertFalse(has_inserted)
        self.assertEqual(1, len(new_postings))
        self.assertEqual(0, len(errors))


class TestBalanceIncompletePostings(cmptest.TestCase):

    def get_incomplete_entry(self, string):
        """Parse a single incomplete entry and convert its CostSpec to a Cost.

        Args:
          string: The input string to parse.
        Returns:
          A pair of (entry, list of errors).
        """
        entries, errors, options_map = parser.parse_string(string, dedent=True)
        self.assertFalse(errors)
        self.assertEqual(1, len(entries))
        (entries_with_lots, errors) = booking_simple.convert_lot_specs_to_lots(entries)
        self.assertEqual(1, len(entries))
        entry = entries_with_lots[0]
        errors = booking_simple.balance_incomplete_postings(entry, options_map)
        return entry, errors

    def test_balance_incomplete_postings__noop(self):
        entry, errors = self.get_incomplete_entry("""
          option "booking_algorithm" "SIMPLE"
          2013-02-23 * "Something"
            Liabilities:CreditCard     -50 USD
            Expenses:Restaurant         50 USD
        """)
        self.assertFalse(errors)
        self.assertEqual(2, len(entry.postings))

    def test_balance_incomplete_postings__fill1(self):
        entry, errors = self.get_incomplete_entry("""
          option "booking_algorithm" "SIMPLE"
          2013-02-23 * "Something"
            Liabilities:CreditCard     -50 USD
            Expenses:Restaurant
        """)
        self.assertFalse(errors)
        self.assertEqual(2, len(entry.postings))
        self.assertEqual(position.get_position(entry.postings[1]),
                         position.from_string('50 USD'))

    def test_balance_incomplete_postings__fill2(self):
        entry, errors = self.get_incomplete_entry("""
          option "booking_algorithm" "SIMPLE"
          2013-02-23 * "Something"
            Liabilities:CreditCard     -50 USD
            Liabilities:CreditCard     -50 CAD
            Expenses:Restaurant
        """)
        self.assertFalse(errors)
        self.assertEqual(4, len(entry.postings))
        self.assertEqual(entry.postings[2].account, 'Expenses:Restaurant')
        self.assertEqual(entry.postings[3].account, 'Expenses:Restaurant')
        self.assertEqual(position.get_position(entry.postings[2]),
                         position.from_string('50 USD'))
        self.assertEqual(position.get_position(entry.postings[3]),
                         position.from_string('50 CAD'))

    def test_balance_incomplete_postings__cost(self):
        entry, errors = self.get_incomplete_entry("""
          option "booking_algorithm" "SIMPLE"
          2013-02-23 * "Something"
            Assets:Invest     10 MSFT {43.23 USD}
            Assets:Cash
        """)
        self.assertFalse(errors)
        self.assertEqual(2, len(entry.postings))
        self.assertEqual(entry.postings[1].account, 'Assets:Cash')
        self.assertEqual(position.get_position(entry.postings[1]),
                         position.from_string('-432.30 USD'))

    def test_balance_incomplete_postings__insert_rounding(self):
        entry, errors = self.get_incomplete_entry("""
          option "booking_algorithm" "SIMPLE"
          option "account_rounding" "RoundingError"

          2013-02-23 * "Something"
            Assets:Invest     1.245 RGAGX {43.23 USD}
            Assets:Cash      -53.82 USD
        """)
        self.assertFalse(errors)
        self.assertEqual(3, len(entry.postings))
        self.assertEqual(entry.postings[2].account, 'Equity:RoundingError')
        self.assertEqual(position.get_position(entry.postings[2]),
                         position.from_string('-0.00135 USD'))

    def test_balance_incomplete_postings__quantum(self):
        entry, errors = self.get_incomplete_entry("""
          option "booking_algorithm" "SIMPLE"
          option "inferred_tolerance_default" "USD:0.01"

          2013-02-23 * "Something"
            Assets:Invest     1.245 RGAGX {43.23 USD}
            Assets:Cash
        """)
        self.assertFalse(errors)
        self.assertEqual(D('-53.82'), entry.postings[1].units.number)

        entry, errors = self.get_incomplete_entry("""
          option "booking_algorithm" "SIMPLE"
          option "inferred_tolerance_default" "USD:0.001"

          2013-02-23 * "Something"
            Assets:Invest     1.245 RGAGX {43.23 USD}
            Assets:Cash
        """)
        self.assertFalse(errors)
        self.assertEqual(D('-53.821'), entry.postings[1].units.number)

    def test_balance_incomplete_postings__rounding_and_quantum(self):
        entry, errors = self.get_incomplete_entry("""
          option "booking_algorithm" "SIMPLE"
          option "account_rounding" "RoundingError"
          option "inferred_tolerance_default" "USD:0.01"

          2013-02-23 * "Something"
            Assets:Invest     1.245 RGAGX {43.23 USD}
            Assets:Cash
        """)
        self.assertFalse(errors)
        self.assertEqual(3, len(entry.postings))
        self.assertEqual(D('-53.82'), entry.postings[1].units.number)
        self.assertEqual('Equity:RoundingError', entry.postings[2].account)
        self.assertEqual(D('-0.00135'), entry.postings[2].units.number)

        entry, errors = self.get_incomplete_entry("""
          option "booking_algorithm" "SIMPLE"
          option "account_rounding" "RoundingError"
          option "inferred_tolerance_default" "USD:0.01"

          2014-05-06 * "Buy mutual fund"
            Assets:Investments:RGXGX       4.27 RGAGX {53.21 USD}
            Assets:Investments:Cash
        """)
        self.assertFalse(errors)
        self.assertEqual(3, len(entry.postings))
        self.assertEqual(D('-227.2100'), entry.postings[1].units.number)
        self.assertEqual('Equity:RoundingError', entry.postings[2].account)
        self.assertEqual(D('0.0033'), entry.postings[2].units.number)

    def test_balance_incomplete_postings__rounding_with_error(self):
        # Here we want to verify that auto-inserting rounding postings does not
        # disable non-balancing transactions. This is rather an important check!
        entries, errors, options_map = loader.load_string("""
          option "booking_algorithm" "SIMPLE"
          option "account_rounding" "RoundingError"

          2000-01-01 open Assets:Investments:MutualFunds:XXX
          2000-01-01 open Assets:Cash:Checking
          2000-01-01 open Equity:RoundingError

          ;; This transaction does not balance.
          2002-02-08 * "Mutual fund purchase"
            Assets:Investments:MutualFunds:XXX    51.031 XXX {97.98 USD}
            Assets:Cash:Checking                -5000.00 USD
        """, dedent=True)
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, 'does not balance')

    @loader.load_doc(expect_errors=True)
    def test_balance_incomplete_postings__superfluous_unneeded(self, entries, errors, _):
        """
          option "booking_algorithm" "SIMPLE"

          2000-01-01 open Assets:Account1
          2000-01-01 open Assets:Account2
          2000-01-01 open Assets:Account3

          2016-04-23 * ""
            Assets:Account1   100.00 USD
            Assets:Account2  -100.00 USD
            Assets:Account3
        """
        # Note: this raises an error because the auto-posting is superfluous.
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, 'Useless auto-posting')
        self.assertEqual(3, len(entries[-1].postings))

    @loader.load_doc()
    def test_balance_incomplete_postings__superfluous_unused(self, entries, errors, _):
        """
          option "booking_algorithm" "SIMPLE"

          2000-01-01 open Assets:Account1
          2000-01-01 open Assets:Account2

          2016-04-23 * ""
            Assets:Account1     0.00 USD
            Assets:Account2
        """
        # This does not raise an error, but it ought to; do this in the full
        # booking method since we're deprecating this code.
        self.assertEqual(1, len(entries[-1].postings))


class TestSimpleBooking(cmptest.TestCase):

    @loader.load_doc()
    def test_simple_booking_algorithm(self, entries, _, options_map):
        """
          option "booking_algorithm" "SIMPLE"

          2013-05-01 open Assets:Bank:Investing
          2013-05-01 open Equity:Opening-Balances

          2013-05-02 *
            Assets:Bank:Investing                 5 HOOL {502 USD}
            Equity:Opening-Balances

          2013-05-02 *
            Assets:Bank:Investing                 5 HOOL {500 # 10 USD}
            Equity:Opening-Balances

          2013-05-02 *
            Assets:Bank:Investing                 5 HOOL {# 2510 USD}
            Equity:Opening-Balances

          2013-05-02 *
            Assets:Bank:Investing                 2510 USD
            Equity:Opening-Balances
        """
        for entry in entries:
            if not isinstance(entry, data.Transaction):
                continue
            self.assertEqual(D('-2510'), entry.postings[-1].units.number)

    @loader.load_doc(expect_errors=True)
    def test_simple_booking_algorithm__invalid(self, _, errors, __):
        """
          option "booking_algorithm" "XXX"
        """
        self.assertEqual(1, len(errors))

    @loader.load_doc(expect_errors=True)
    def test_simple_booking_algorithm__issue139(self, _, errors, __):
        """
          option "booking_algorithm" "SIMPLE"
          plugin "beancount.plugins.auto_accounts"

          2016-08-06 * "Test"
            Expenses:Other                          57.30 BRL @@ 16.18 EUR
            Expenses:Food                             21.90 BRL @@ 6.18
            Assets:Something
        """
        self.assertRegex(errors[0].message,
                         'Missing number or currency on price not handled')

    @loader.load_doc(expect_errors=True)
    def test_no_cost_is_rejected(self, _, errors, __):
        """
          option "booking_algorithm" "SIMPLE"
          plugin "beancount.plugins.auto_accounts"

          2014-10-15 * "buy widgets"
            Assets:Inventory     10 WIDGET {} ;; Not supported.
            Assets:Cash         -80 GBP
        """
        # See https://groups.google.com/d/msg/beancount/9NBcT-SXZMQ/sy7Z3dIdBQAJ
        self.assertRegex(errors[0].message,
                         'Cost syntax not supported')



__copyright__ = "Copyright (C) 2013-2017  Martin Blais"
__license__ = "GNU GPLv2"

import warnings
warnings.warn("beancount.ops.prices is deprecated; has moved to beancount.core.prices")

from beancount.core.prices import *  # pylint: disable=wildcard-import,unused-wildcard-import
